import difflib
import os
from pathlib import Path
from typing import List

from loguru import logger


def get_file(file_path):
    with open(file_path) as f:
        data = f.read()
    return data


def get_diffing_lines(old_lines, new_lines) -> List[int]:
    diffs = difflib.Differ().compare(old_lines, new_lines)
    lin_num = 0
    diffing_lines_num = []
    diffing_lines = []
    for line in diffs:
        code = line[:2]
        if code in ("  ", "+ "):
            lin_num += 1
        if code == "+ ":
            diffing_lines.append(line)
            diffing_lines_num.append(lin_num)
    logger.info(f"Diff. lines numbers: {diffing_lines_num}")
    logger.info(f"Diff. lines: {diffing_lines}")
    return diffing_lines_num


def get_file_path_from_notebook_path(notebook_path):
    notebook_path = Path(notebook_path)
    return notebook_path.parent / f"{notebook_path.stem}.py"


def get_notebook_path_from_file_path(file_path):
    file_path = Path(file_path)
    return file_path.parent / f"{file_path.stem}.ipynb"


def get_file_update_timestamp(file_path):
    statbuf = os.stat(file_path)
    timestamp = statbuf.st_mtime
    return timestamp



import json
from sys import stderr
from time import sleep
from typing import List, Dict, Optional

import jupytext
from loguru import logger
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, WebDriverException
from selenium.webdriver.chrome.options import Options
from watchdog.events import FileSystemEventHandler

from jure.events import cells_changed_event, EventType
from jure.utils import (
    get_notebook_path_from_file_path,
    get_file_update_timestamp,
    get_file,
)

logger.remove()
logger.add(stderr, format="{time: HH:mm:SSSS} {level} {message}", level="INFO")


class BaseHandler:
    def handle(self, event):
        print(event)

    def shutdown(self):
        pass


class SeleniumHandler(BaseHandler):
    def __init__(self, page):
        options = Options()
        options.add_experimental_option("excludeSwitches", ["enable-automation"])
        options.add_experimental_option("useAutomationExtension", False)
        options.set_capability("unhandledPromptBehaviour", "accept")
        options.add_argument("--disable-popup-blocking")
        self.driver = webdriver.Chrome(options=options)
        self.driver.execute_script("window.onbeforeunload = null;")
        try:
            self.driver.get(page)
            self.driver.execute_script("Jupyter.notebook.set_autosave_interval(0);")
        except WebDriverException as e:
            msg = f"""Unable to load Jupyter Notebook.
                      Make sure that Jupyter Notebook is available on page {page}"""
            raise RuntimeError(msg)

    def handle(self, event):
        if event["type"] == EventType.RELOAD_PAGE:
            self._refresh_page()
        if event["type"] == EventType.HANDLE_CHANGED_CELLS:
            self._scroll_to_cell(event["cell_to_scroll"])
            self._execute_cells(event["changed_cells"])

    def shutdown(self):
        self.driver.close()

    def _refresh_page(self):
        self.check_popup()
        sleep(0.1)
        self.driver.refresh()
        self.driver.execute_script("Jupyter.notebook.set_autosave_interval(0);")

    def _scroll_to_cell(self, cell_num: Optional[int]):
        self.check_popup()
        if cell_num is not None:
            logger.info(f"Scrolling to cell {cell_num}")
            self.driver.execute_script(
                f"""Jupyter.notebook.scroll_to_cell({cell_num});"""
            )

    def _execute_cells(self, changed_cells: List[Dict]):
        logger.info(f"Executing")
        if changed_cells:
            formatted_cells = [json.dumps(c, ensure_ascii=False) for c in changed_cells]
            update_script = f"""let data = [{','.join(formatted_cells)}];
data.forEach(function (cell) {{
    Jupyter.notebook.get_cell(cell.index).set_text(cell.content)
}});
"""
            try:
                self.driver.execute_script(update_script)
            except:
                self._refresh_page()
            exec_script = f"""
                Jupyter.notebook.execute_cells([{','.join([c["index"] for c in changed_cells])}]);
                """
            self.driver.execute_script(exec_script)

    def check_popup(self):
        try:
            self.driver.switch_to.alert.accept()
        except:
            pass
        reload_btn = None
        try:
            notebook_changed_modal = self.driver.find_element_by_css_selector(
                "body.modal-open"
            )
            reload_btn = notebook_changed_modal.find_element_by_class_name(
                "btn-warning"
            )
        except NoSuchElementException:
            pass
        if reload_btn is not None:
            reload_btn.click()
            try:
                self.driver.switch_to.alert.accept()
            except:
                pass


class WatchdogHandler(FileSystemEventHandler):
    def __init__(self, file_path, handler):
        self.file_path = file_path
        self.notebook_path = get_notebook_path_from_file_path(file_path)
        self.prev_cells = jupytext.reads(get_file(self.file_path), "py:percent")[
            "cells"
        ]
        self.prev_update_timestamp = 0
        self.handler = handler

    def on_modified(self, event):
        file_update_timestamp = get_file_update_timestamp(self.file_path)
        notebook_update_timestamp = get_file_update_timestamp(self.notebook_path)
        if self.should_reload(file_update_timestamp, notebook_update_timestamp):
            logger.info("Updating notebook text")
            try:
                cells = jupytext.reads(get_file(self.file_path), "py:percent")["cells"]
                diffing_cells = [
                    i
                    for i, cell in enumerate(cells)
                    if i >= len(self.prev_cells)
                    or self.prev_cells[i]["source"] != cell["source"]
                ]
                self.prev_cells = cells
                cell_to_scroll = None
                if diffing_cells:
                    cell_to_scroll = diffing_cells[-1]
                cell_info = [
                    {"index": str(cell_index), "content": cells[cell_index]["source"]}
                    for cell_index in sorted(diffing_cells)
                ]
                new_event = cells_changed_event(
                    cell_to_scroll=cell_to_scroll, diffing_cells=cell_info
                )
                logger.info(f"{new_event['type']}. Changed cells: {diffing_cells}")
                self.handler.handle(new_event)
            except Exception as e:
                logger.exception(e)
        self.prev_update_timestamp = file_update_timestamp

    def should_reload(self, file_update_timestamp, notebook_update_timestamp):
        logger.info(f"update ts file:{file_update_timestamp}, nb:{notebook_update_timestamp}")
        if (file_update_timestamp - self.prev_update_timestamp) > 0.5:
            return True
        else:
            logger.info("File was recently updated: no change")
        return False


class AngryHandler(BaseHandler):
    def handle(self, event):
        raise RuntimeError("You did a mistake...")



from enum import Enum


class EventType(Enum):
    RELOAD_PAGE = 1
    HANDLE_CHANGED_CELLS = 2


reload_event = dict(type=EventType.RELOAD_PAGE)


def cells_changed_event(diffing_cells, cell_to_scroll):
    return {"type": EventType.HANDLE_CHANGED_CELLS,
            "changed_cells": diffing_cells,
            "cell_to_scroll": cell_to_scroll}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jure']

package_data = \
{'': ['*']}

install_requires = \
['jupytext>=1.5.2,<2.0.0',
 'loguru>=0.5.2,<0.6.0',
 'pytest>=6.0.1,<7.0.0',
 'selenium>=3.141.0,<4.0.0',
 'watchdog>=0.10.3,<0.11.0']

entry_points = \
{'console_scripts': ['jure = jure.main:main']}

setup_kwargs = {
    'name': 'jure',
    'version': '0.3.1',
    'description': 'An utility that extends Jupytext. Allows to auto-refresh browser when source file is changed.',
    'long_description': "# Jupyter Browser Reload\n\nFlow-saving tool that automatically reloads Jupyter Notebook in a browser\nwhen its source .py file is changed and executes all changed cells.\n\nIt uses\n* [Jupytext](https://github.com/mwouts/jupytext) - to synchronize .ipynb and .py files\n* [Watchdog](https://github.com/gorakhargosh/watchdog) - to watch for .py file changes\n* [Selenium](https://github.com/SeleniumHQ/selenium) - to have a full control over a browser with opened Jupyter Notebook \n\n## Why Jure\nJupytext is a great tool that for instance allows user to benefit from static code analysis of Jupyter Notebooks. However I always struggled with this workflow: after each edit of .py file I needed to manually reload browser and execute all changed cells.\n\n![standard](assets/standard.gif)\n\nJure automatically reloads browser on each .py file change, so it would instantly show actual notebook content. Additionally it scrolls to last changed cell and executes all cells that were changed. \n \n![with jure](assets/with_jure.gif)\n\n\n## Installation\nThe most non-trivial part is install ChromeDriver on you computer, here's [sample link for Ubuntu](https://tecadmin.net/setup-selenium-chromedriver-on-ubuntu/) for reference.\nAfter that it's simply\n```\npip install jure\n```\n\n## Usage\nFirst, you need to sync your .ipynb file with .py with Jupytext, see [official manual](https://github.com/mwouts/jupytext#using-jupytext).\n\nThen launch a Jupyter Notebook server (`jupyter notebook ...`).\n\nThen run\n```bash\njure --token=[TOKEN] --jupyter_root_dir=[ROOT_DIR] --notebook_path=[NOTEBOOK_PATH]\n```\nWhere `[TOKEN]` as an access token which is required to access Jupyter Notebook from browser, `[ROOT_DIR]` is a path to directory from which `jupyter notebook ...` command was executed and `[NOTEBOOK_PATH]` is a path to the notebook .ipynb file you'll work with.\n\n## Limitations\nThis is an experimental and unstable product, any issues, suggestions, feature requests and PRs are appreciated. Current problems:\n\n* Only Google Chrome web browser is supported\n* Selenium might be inconvenient\n* No password auth / remote notebook hosts\n* In some rare cases user needs to reload browser tab manually\n* For large notebooks page reload might be too slow (tough one)",
    'author': 'Dmitry Lipin',
    'author_email': 'd.lipin@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
